/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef CREATE_H
#define CREATE_H

#include "Progress.hh"
#include "Polygon3D.hh"
#include <fstream>
#include <memory>
#include <string>

namespace SharpConstruct
{
	class Mesh;

	namespace Create
	{
		class File : public Stepable
		{
		public:
			static const bool RequiresInitialization = false;

			File( const std::string& );

			bool Load( Mesh& );
			bool operator()( Mesh& );

			virtual double Progress() const;
			virtual const std::string Text() const;

			bool Step();
		private:
			class LoadOBJ : public Stepable
			{
			public:
				LoadOBJ( Mesh&, const std::string& );
				
				virtual double Progress() const;
				virtual const std::string Text() const;
				virtual bool Step();
			private:
				void load_line_();
				void count_lines_( const std::string&);

				Mesh& mesh_;
				std::ifstream file_;
				std::vector< Polygon3D > polygons_;
				unsigned current_polygon_;
				bool colored_;
				int line_count_, line_;
			};
			class LoadLWO : public Stepable
			{
			public:
				LoadLWO( Mesh&, const std::string& );
				
				virtual double Progress() const;
				virtual bool Step();
			private:
				typedef std::vector< char > Data;
				
				/* Reads n characters from file_, and reverses byte
				   order when necessary.
				 */
				Data read_chars_( const unsigned n );
				
				unsigned short read_u2_();
				unsigned int read_u4_();
				int read_sx_(); // Read an X-length (2 or 4 byte) number
				float read_float_();
				std::string read_string_( const unsigned );
				bool compare_chunk_( const unsigned, const std::string& );

				Mesh& mesh_;
				std::ifstream file_;
				unsigned offset_;
			};

			const std::string file_;
			std::auto_ptr< Stepable > loader_;
		};

		class Cube
		{
		public:
			static const bool RequiresInitialization = true;

			Cube( Progress&, const unsigned, const unsigned,
			      const unsigned );

			bool operator()( Mesh& );
		private:
			void create_vertices_();
			void create_back_vertices_();
			void create_top_vertices_();
			void create_front_vertices_();
			void create_bottom_vertices_();
			void create_left_vertices_();
			void create_right_vertices_();

			void create_polygons_() const;
			void create_main_polygon_strip_() const;
			void connect_back_to_bottom_() const;
			void create_left_and_right_polygons_() const;
			void create_vertical_connections_() const;
			void create_horizontal_connections_() const;
			void create_left_corners_() const;
			void create_right_corners_() const;

			void add_quad_( const unsigned, const unsigned,
					const unsigned, const unsigned ) const;

			Progress& progress_;

			const unsigned rows_, cols_, stacks_;
			unsigned top_start_, front_start_, bottom_start_;
			unsigned left_start_, right_start_, right_end_;
		};

		class Sphere
		{
		public:
			static const bool RequiresInitialization = true;

			Sphere( Progress&, const unsigned );

			bool operator()( Mesh& ) const;
		private:
			Progress& progress_;

			const unsigned divisions_;
		};

		class Plane
		{
		public:
			static const bool RequiresInitialization = true;

			Plane( const unsigned, const unsigned );
			
			bool operator()( Mesh& ) const;
		private:
			const unsigned rows_, columns_;
		};
	}
}

#endif
