/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Create.hh"
#include "Mesh.hh"
#include "Utilities.hh"

using SharpConstruct::Create::File;
using SharpConstruct::Mesh;

File::File( const std::string& file )
: file_( file )
{}

bool File::Load( Mesh& mesh )
{
	const std::string extension( ExtractExtension( file_ ) );

	if( extension == "OBJ" )
	{
		loader_.reset( new LoadOBJ( mesh, file_ ) );
		return true;
	}
	else if( extension == "LWO" )
	{
		loader_.reset( new LoadLWO( mesh, file_ ) );
		return true;
	}
	else
		return false;
}

double File::Progress() const
{
	return loader_->Progress();
}

const std::string File::Text() const
{
	return loader_->Text();
}

bool File::operator()( Mesh& mesh )
{
	mesh.SetFileName( file_ );

	return false;
}

bool File::Step()
{
	if( !loader_.get() )
		return false;
	return loader_->Step();
}

File::LoadOBJ::LoadOBJ( Mesh& m, const std::string& filename )
: mesh_( m ), file_( filename.c_str(), std::ios::in ),
  current_polygon_( 0 ),
  colored_( false ), line_( 0 )
{
	count_lines_( filename );
}

double File::LoadOBJ::Progress() const
{
	if( file_ )
		return line_ * 1.0 / line_count_;
	else
		return current_polygon_ * 1.0 / polygons_.size();
}

const std::string File::LoadOBJ::Text() const
{
	if( file_ )
		return "Loading...";
	else
		return "Creating polygons...";
}

bool File::LoadOBJ::Step()
{
	if( file_ )
	{
		if( !file_.eof() )
		{
			load_line_();
			++line_;
		}
		else
			file_.close();
		return true;
	}
	else // Now start converting Polygon3Ds to tris and quads
	{
		if( current_polygon_ < polygons_.size() )
		{
			const Polygon3D& poly( polygons_[ current_polygon_ ] );

			if( poly.VertexIndices.size() == 3 )
			{
				mesh_.Triangles().push_back( Triangle( poly.VertexIndices[ 2 ],
								       poly.VertexIndices[ 1 ],
								       poly.VertexIndices[ 0 ] ) );
			}
			else if( poly.VertexIndices.size() == 4 )
			{
				mesh_.Quads().push_back( Quad( poly.VertexIndices[ 3 ],
							       poly.VertexIndices[ 2 ],
							       poly.VertexIndices[ 1 ],
							       poly.VertexIndices[ 0 ] ) );
			}
			else
			{
				for( unsigned j = 1; j < poly.VertexIndices.size() - 1; ++j )
				{
					mesh_.Triangles().push_back( Triangle( poly.VertexIndices[ j + 1 ],
									       poly.VertexIndices[ j ],
									       poly.VertexIndices[ 0 ] ) );
				}
			}
			
			++current_polygon_;
			return true;
		}

		// TC: Added to fix the crash in preview
		mesh_.ResetVisibles();

		// Do this in one chunk
		if( !colored_ )
		{
			mesh_.VertexColors().resize( mesh_.VertexLocations().size() );
			mesh_.Flood( Color( 1, 1, 1 ), 1 );
		}

		return false;
	}
}

void File::LoadOBJ::load_line_()
{
	std::string token, subtoken;
	char buffer[ 256 ];
	std::istringstream line;

	file_.getline( buffer, 256 );
	line.str( buffer );
	line >> token;
	if( token == "v" )
	{
		float x, y, z;
		line >> x >> y >> z;
		mesh_.VertexLocations().push_back( Point3D( x, y, z ) );
	}
	else if( token == "f" )
	{
		Polygon3D poly;
		while( line >> token )
		{
			int loc = token.find( '/' );
			subtoken = token.substr( 0, loc );
			poly.VertexIndices.push_back( ToInt( subtoken ) - 1 );
		}
		polygons_.push_back( poly );
	}
	else if( token == "#vertexcolor" )
	{
		unsigned ndx;
		line >> ndx;
		mesh_.VertexColors().push_back( SharpConstruct::Color() );
		float r, g, b;
		line >> r >> g >> b;
		mesh_.VertexColors()[ ndx ] = Color( r, g, b );
		colored_ = true;
	}
}

void File::LoadOBJ::count_lines_( const std::string& filename )
{
	line_count_ = 0;

	std::ifstream file( filename.c_str(), std::ios::in );

	if( file )
	{
		char ch;
		while( !file.eof() )
		{
			file.get( ch );
			if( ch == '\n' )
				++line_count_;
		}
	}
}

#if 0
void File::load_obj_( Mesh& mesh ) const
{
	std::ifstream file( file_.c_str(), std::ios::in );
	//std::istringstream line;
	std::string token, subtoken;
	char buffer[ 256 ];
	unsigned polygon_count = 0;
	unsigned ndx;
	bool hascolor = false;

	std::vector< Polygon3D > polygons;

	while( !file.eof() )
	{
		std::istringstream line;
		file.getline( buffer, 256 );
		line.str( buffer );
		line >> token;
		if( token == "v" )
		{
			float x, y, z;
			line >> x >> y >> z;
			mesh.VertexLocations().push_back( Optimized::Point3D( x, y, z ) );
		}
		else if( token == "f" )
		{
			polygons.resize( polygon_count + 1 );
			while( line >> token )
			{
				int loc = token.find( '/' );
				subtoken = token.substr( 0, loc );
				polygons[ polygon_count ].VertexIndices.push_back( ToInt( subtoken ) - 1 );
			}
			polygon_count++;
		}
		else if( token == "#vertexcolor" )
		{
			line >> ndx;
			mesh.VertexColors().push_back( SharpConstruct::Color() );
			line >> mesh.VertexColors()[ ndx ].Red >> mesh.VertexColors()[ ndx ].Green >>
			        mesh.VertexColors()[ ndx ].Blue;
			hascolor = true;
		}
	}
	file.close();

	for( unsigned i = 0; i < polygons.size(); ++i )
	{
		if( polygons[ i ].VertexIndices.size() == 3 )
		{
			mesh.Triangles().push_back( Triangle( polygons[ i ].VertexIndices[ 2 ],
							      polygons[ i ].VertexIndices[ 1 ],
							      polygons[ i ].VertexIndices[ 0 ] ) );
		}
		else if( polygons[ i ].VertexIndices.size() == 4 )
		{
			mesh.Quads().push_back( Quad( polygons[ i ].VertexIndices[ 3 ],
						      polygons[ i ].VertexIndices[ 2 ],
						      polygons[ i ].VertexIndices[ 1 ],
						      polygons[ i ].VertexIndices[ 0 ] ) );
		}
		else
		{
			for( unsigned j = 1; j < polygons[ i ].VertexIndices.size() - 1; ++j )
			{
				mesh.Triangles().push_back( Triangle( polygons[ i ].VertexIndices[ j + 1 ],
								      polygons[ i ].VertexIndices[ j ],
								      polygons[ i ].VertexIndices[ 0 ] ) );
			}
		}
	}
	
	/*_recalculate_vertex_users();
	  RecalculateAllNormals();*/
	
	if( !hascolor )
	{
		mesh.VertexColors().resize( mesh.VertexLocations().size() );
		mesh.Flood( Color( 1, 1, 1 ), 1 );
	}
	
	//_saved = true;
}
#endif
