/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Brush.hh"
#include "Grid.hh"
#include "MeshHistory.hh"
#include "Point.hh"
#include <cmath>

#ifdef _MSC_VER
#include "win32.h"
#endif

using SharpConstruct::Grid;
using SharpConstruct::Brush::EditData;
using SharpConstruct::Point3D;

void Grid::SetDivision( const unsigned d )
{
	division_ = d;

	MeshHistory::Instance().GetCurrentMesh().UpdateBounds();
	UpdateCells();
}

const std::vector< unsigned >& Grid::operator()( const unsigned r,
						 const unsigned c,
						 const unsigned s ) const
{
	return cells_[ r * ( division_ * division_ ) + c * division_ + s ];
}
const std::vector< unsigned >& Grid::operator()( const Coordinate& c ) const
{
	return cells_[ c[ 0 ] * ( division_ * division_ ) + c[ 1 ] * division_ + c[ 2 ] ];
}
std::vector< unsigned >& Grid::operator()( const Coordinate& c )
{
	return cells_[ c[ 0 ] * ( division_ * division_ ) + c[ 1 ] * division_ + c[ 2 ] ];
}
std::vector< unsigned >& Grid::operator()( const unsigned r,
					   const unsigned c,
					   const unsigned s )
{
	return cells_[ r * ( division_ * division_ ) + c * division_ + s ];
}

void Grid::FindActiveCells( const EditData& e )
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );
	const Point3D& center( mesh.Bounds().Center() );
	const Point3D& sel( e.Center );
	const float diam( mesh.Bounds().Diameter() );
	const float mrad( diam / 2.0f );
	const float radius( mesh.SelectionRadius() );
	const float minx( center.X() - mrad );
	const float miny( center.Y() - mrad );
	const float minz( center.Z() - mrad );

	active_cells_.clear();

	const Range< float > xrng( sel.X(), radius );
	const Range< float > yrng( sel.Y(), radius );
	const Range< float > zrng( sel.Z(), radius );

	/* Note that these three loops are looping
	   through the cell vertices, not the cells. */
	for( unsigned r = 1; r < division_; ++r )
	{
		const float x( cell_to_modelspace_( division_, diam, minx, r ) );
		for( unsigned c = 1; c < division_; ++c )
		{
			const float y( cell_to_modelspace_( division_, diam, miny, c ) );
			for( unsigned s = 1; s < division_; ++s )
			{
				Coordinate coord( r, c, s );
				const float z( cell_to_modelspace_( division_, diam, minz, s ) );

				if( xrng( x ) &&
				    yrng( y ) &&
				    zrng( z ) )
				{
					make_cell_neighbors_active_( coord );
				}
			}
		}
	}
	
	make_cell_neighbors_active_( FindPoint( sel ) );
}

Grid::Coordinate Grid::FindPoint( const Point3D& v ) const
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );
	const float diam( mesh.Bounds().Diameter() );
	const float radius( diam / 2.0f );
	const float minx( mesh.Bounds().Center().X() - radius );
	const float miny( mesh.Bounds().Center().Y() - radius );
	const float minz( mesh.Bounds().Center().Z() - radius );
	return Coordinate( convert_range_( division_, diam, minx, v.X() ),
			   convert_range_( division_, diam, miny, v.Y() ),
			   convert_range_( division_, diam, minz, v.Z() ) );
}

bool Grid::Valid( const Coordinate& c ) const
{
	return c[ 0 ] < division_ &&
	       c[ 1 ] < division_ &&
	       c[ 2 ] < division_;
}

const Grid::ActiveCellType& Grid::ActiveCells() const
{
	return active_cells_;
}
Grid::ActiveCellType& Grid::ActiveCells()
{
	return active_cells_;
}

void Grid::UpdateCells()
{
	const Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	// Clear it to get rid of old cells
	cells_.clear();
	cells_.resize( division_ * division_ * division_ );

	const float diam( mesh.Bounds().Diameter() );
	const float radius( diam / 2.0f );
	const float minx( mesh.Bounds().Center().X() - radius );
	const float miny( mesh.Bounds().Center().Y() - radius );
	const float minz( mesh.Bounds().Center().Z() - radius );
	const unsigned size( mesh.VertexLocations().size() );
	for( unsigned i = 0; i < size; ++i )
	{
		const Point3D& v( mesh.VertexLocations()[ i ] );

		operator()( convert_range_( division_, diam, minx, v.X() ),
			    convert_range_( division_, diam, miny, v.Y() ),
			    convert_range_( division_, diam, minz, v.Z() ) ).push_back( i );
	}

	active_cells_.clear();
}

void Grid::make_cell_neighbors_active_( const Coordinate c )
{
	for( int dx = -1; dx <= 1; ++dx )
		for( int dy = -1; dy <= 1; ++dy )
			for( int dz = -1; dz <= 1; ++dz )
				make_cell_active_( c.Offset( dx, dy, dz ) );
}

void Grid::make_cell_active_( const Coordinate c )
{
	if( Valid( c ) )
		active_cells_.insert( &operator()( c ) );
}

unsigned Grid::convert_range_( const unsigned len, const float diam,
			       const float min, const float pos ) const
{
	const float r( ( ( len - 1 ) * ( pos - min ) ) / diam );
	if( r < 0 )
		return 0;
	if( r > len )
		return len;
	else
		return lroundf( r );
}

float Grid::cell_to_modelspace_( const unsigned len, const float diam,
				 const float min, const unsigned cell ) const
{
	return ( diam * cell ) / ( len - 1 ) + min;
}
