/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "ModelDisplay.hh"
#include "Progress.hh"
#include <gdkmm/cursor.h>
#include <iostream>

#include <GL/glu.h>

#ifdef WIN32
#include "win32.h"
#endif

using SharpConstruct::ModelDisplay;
using SharpConstruct::Mesh;

ModelDisplay::ModelDisplay()
: arc_ball_( 0.8 ), zoom_( 1 )
{
	add_events( Gdk::POINTER_MOTION_MASK |
		    Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK );
	set_extension_events( Gdk::EXTENSION_EVENTS_CURSOR );

	Glib::RefPtr<Gdk::GL::Config> glconfig;

	glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB    |
					   Gdk::GL::MODE_DEPTH  |
					   Gdk::GL::MODE_DOUBLE);
	if (!glconfig)
	{
		glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB   |
						   Gdk::GL::MODE_DEPTH);
		if (!glconfig)
		{
			std::cerr << "*** Cannot find any OpenGL-capable visual.\n";
			std::exit( 1 );
		}
	}
	
	set_gl_capability( glconfig );

	show();
}

void ModelDisplay::SetMesh( Mesh& m )
{
	mesh_ = &m;
}

void ModelDisplay::on_realize()
{
	Gtk::DrawingArea::on_realize();
	
	Glib::RefPtr<Gdk::GL::Window> glwindow = get_gl_window();
	
	if (!glwindow->gl_begin(get_gl_context()))
		return;

	glEnable( GL_BLEND );
	glEnable( GL_COLOR_MATERIAL );
	glEnable( GL_DEPTH_TEST );

	// Enable lighting
	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );
	const float ambient[] = { 0.15, 0.15, 0.15, 1 };
	const float diffuse[] = { 0.6, 0.6, 0.6, 1 };
	const float specular[] = { 0, 0, 0, 0 };
	const float position[] = { 0, 0, -1, 0 };
	glLightfv( GL_LIGHT0, GL_AMBIENT, ambient );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse );
	glLightfv( GL_LIGHT0, GL_SPECULAR, specular );
	glLightfv( GL_LIGHT0, GL_POSITION, position );

	//DisplayOptions().SetGraphicsMode();
	glout_ = new VBO;

	get_window()->set_cursor( Gdk::Cursor( Gdk::EXCHANGE ) );
	
	glwindow->gl_end();
}

bool ModelDisplay::on_configure_event( GdkEventConfigure* e )
{
	if ( !get_gl_window()->gl_begin( get_gl_context() ) )
		return false;

	set_viewport_();

	get_gl_window()->gl_end();
 
	return true;
}

bool ModelDisplay::on_expose_event( GdkEventExpose* e )
{
	Glib::RefPtr< Gdk::GL::Window > glwindow = get_gl_window();

	if( !get_gl_window()->gl_begin( get_gl_context() ) )
		return false;

	glClearColor( 0.3, 0.3, 0.3, 1 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	set_viewport_();
	orientate_();

	glEnable( GL_LIGHTING );
	glShadeModel( GL_SMOOTH );
	glDisable( GL_CULL_FACE );
	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );

	glout_->Set( *mesh_, true, true, true );
	glout_->Draw();

	// Swap buffers
	if (glwindow->is_double_buffered())
		glwindow->swap_buffers();
	else
		glFlush();
	
	get_gl_window()->gl_end();

	return true;
}

bool ModelDisplay::on_button_press_event( GdkEventButton* e )
{
	if( e->button == 1 || e->button == 3 )
	{
		double xnorm = last_mouse_location_.X();
		double ynorm = last_mouse_location_.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, get_width(), get_height() );
		arc_ball_.BeginDrag( xnorm, ynorm );
	}
	return true;
}

bool ModelDisplay::on_motion_notify_event( GdkEventMotion* e )
{
	if( e->state& GDK_BUTTON1_MASK || e->state & GDK_BUTTON3_MASK )
	{
		double xnorm = last_mouse_location_.X();
		double ynorm = last_mouse_location_.Y();
		ArcBall::NormalizeCoordinates( xnorm, ynorm, get_width(), get_height() );
		arc_ball_.Drag( xnorm, ynorm );
		queue_draw();
	}

	last_mouse_location_ = Math::Point2D< int >( lroundf( e->x ), lroundf( e->y ) );

	return true;
}

bool ModelDisplay::on_button_release_event( GdkEventButton* e )
{
	return true;
}

bool ModelDisplay::on_scroll_event( GdkEventScroll* e )
{
	Gtk::Widget::on_scroll_event( e );

	const float zoom_delta( 0.1 );
	if( e->direction == GDK_SCROLL_UP )
		zoom_ -= zoom_delta;
	else if( e->direction == GDK_SCROLL_DOWN )
		zoom_ += zoom_delta;

	if( zoom_ < 0.1 )
		zoom_ = 0.1;

	queue_draw();

	return true;
}

void ModelDisplay::set_viewport_()
{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	glViewport( 0, 0, get_width(), get_height() );
	float ratio( get_width() );
	if( get_height() )
		ratio /= get_height();
	gluPerspective( 45 * zoom_, ratio, 0.1f, 50 );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

void ModelDisplay::orientate_()
{
	double trackball_matrix[ 16 ];
	unsigned col, row, index;

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0, 0, mesh_->Bounds().Diameter(), mesh_->Bounds().Center().X(),
	           mesh_->Bounds().Center().Y(), mesh_->Bounds().Center().Z(),
		   0, 1, 0 );

	const ArcBall::Transform* rot;

	rot = arc_ball_.Transformation();

	for( col = 0, index = 0; col < 4; col++ )
	{
		for( row = 0; row < 4; row++ )
			trackball_matrix[ index++ ] = ( *rot )[ row ][ col ];
	}
  
	glMultMatrixd( trackball_matrix );

	const float amt( 100.0f );
	glTranslatef( mesh_->RightNormal().X() * ( location_.X() / amt ),
		      mesh_->RightNormal().Y() * ( location_.X() / amt ) ,
		      mesh_->RightNormal().Z() * ( location_.X() / amt ) );
	glTranslatef( mesh_->UpNormal().X() * ( location_.Y() / amt ),
		      mesh_->UpNormal().Y() * ( location_.Y() / amt ),
		      mesh_->UpNormal().Z() * ( location_.Y() / amt ) );
}
