/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Save.hh"

using SharpConstruct::SaveLWO;
using SharpConstruct::Mesh;

SaveLWO::SaveLWO( Mesh& mesh, const std::string& filename )
: mesh_( mesh ), file_( filename.c_str(), std::ios::out | std::ios::binary ),
  length_( 0 )
{
	// Header
	write_chunk_( "FORM" );
	const std::streampos pos_formsize( file_.tellp() );
	write_< unsigned >( 0 ); // Must be replaced at end with length_
	write_chunk_( "LWO2" );

	// Fluff and nonsense
	write_chunk_( "TAGS" );
	write_< unsigned >( 8 );
	write_string_( "Default" );
	write_chunk_( "LAYR" );
	write_< unsigned >( 18 );
	write_< unsigned >( 0 );
	write_< float >( 0 );
	write_< float >( 0 );
	write_< float >( 0 );
	write_< unsigned short >( 0 );

	// Vertices
	write_chunk_( "PNTS" );
	write_< unsigned >( 3 * 4 * mesh_.VertexLocations().size() );
	for( unsigned i = 0; i < mesh_.VertexLocations().size(); ++i )
	{
		const Point3D& v( mesh_.VertexLocations()[ i ] );
		write_< float >( v.X() );
		write_< float >( v.Z() );
		write_< float >( v.Y() );
	}

	// Polygons
	write_chunk_( "POLS" );
	const std::streampos pos_polssize( file_.tellp() );
	write_< unsigned >( 0 ); // Again, must be replaced later
	const unsigned pols_length( length_ );
	write_chunk_( "FACE" );

	for( unsigned i = 0; i < mesh_.Triangles().size(); ++i )
	{
		write_< unsigned short >( 3 );
		const Triangle& p( mesh_.Triangles()[ i ] );
		for( unsigned j = 0; j < 3; ++j )
			write_ux_( p.V[ j ] );
	}

	for( unsigned i = 0; i < mesh_.Quads().size(); ++i )
	{
		write_< unsigned short >( 4 );
		const Quad& p( mesh_.Quads()[ i ] );
		for( unsigned j = 0; j < 4; ++j )
			write_ux_( p.V[ j ] );
	}

	const unsigned final_length( length_ );

	file_.seekp( pos_formsize );
	write_< unsigned >( final_length - 8 );
	file_.seekp( pos_polssize );
	write_< unsigned >( final_length - pols_length );
}

void SaveLWO::write_data_( Data data )
{
	// Reverse data on little-endian machines
	if( G_BYTE_ORDER == G_LITTLE_ENDIAN )
		reverse( data.begin(), data.end() );

	// Write data to file
	for( unsigned i = 0; i < data.size(); ++i )
		file_.put( data[ i ] );

	length_ += data.size();
}

void SaveLWO::write_ux_( const unsigned t )
{
	if( t < 0xFF00 )
		write_< unsigned short >( t );
	else
	{
		Data data( 4 );
		for( unsigned i = 0; i < 4; ++i )
			data[ i ] = ( ( char* )( &t ) )[ i ];
		data[ 3 ] = 0xFF;
		write_data_( data );
	}
}

void SaveLWO::write_chunk_( const std::string& s )
{
	Data data;
	for( int i = 0; i < 4; ++i )
		data.push_back( s[ 3 - i ] );
	write_data_( data );
}

void SaveLWO::write_string_( const std::string& s )
{
	Data data;
	data.push_back( '\0' );
	for( int i = 0; i < static_cast< signed >( s.size() ); ++i )
		data.push_back( s[ s.size() - 1 - i ] );
	write_data_( data );
}
